# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

# Data format versioning
__report_version__ = '1.1.0'

import time
import json
import hashlib

from collections import deque
from base64 import b64encode

PROC_CREATE = 'create'
MEM_ALLOC = 'mem_alloc'
MEM_WRITE = 'mem_write'
MEM_READ = 'mem_read'
MEM_PROTECT = 'mem_protect'
THREAD_INJECT = 'thread_inject'


class ProfileError(Exception):
    pass


class MemAccess(object):
    """
    Represents a symbolicated chunk of memory that can be tracked
    """
    def __init__(self, base=None, sym=None, size=0):
        self.base = base
        self.size = size
        self.sym = sym
        self.reads = 0
        self.writes = 0
        self.execs = 0


class Run(object):
    """
    This class represents the basic execution primative for the emulation engine
    A "run" can represent any form of execution: a thread, a callback, an exported function,
    or even a child process.
    """
    def __init__(self):
        self.instr_cnt = 0
        self.ret_val = None
        self.apis = []
        self.sym_access = {}
        self.network = {'dns': [], 'traffic': []}
        self.file_access = []
        self.dropped_files = []
        self.registry_access = []
        self.process_events = []
        self.mem_access = {}
        self.dyn_code = {'mmap': [], 'base_addrs': set()}
        self.process_context = None
        self.thread = None
        self.unique_apis = []
        self.api_hash = hashlib.sha256()
        self.handled_exceptions = []
        self.stack = None
        self.api_callbacks = []
        self.exec_cache = deque(maxlen=4)
        self.read_cache = deque(maxlen=4)
        self.write_cache = deque(maxlen=4)

        self.args = None
        self.start_addr = None
        self.type = None
        self.error = {}
        self.num_apis = 0

    def get_api_count(self):
        """
        Get the number of APIs that were called during the run
        """
        return self.num_apis


class Profiler(object):
    """
    The profiler class exists to generate an execution report
    for all runs that occur within a binary emulation.
    """
    def __init__(self):
        super(Profiler, self).__init__()

        self.start_time = 0
        self.strings = {'ansi': [], 'unicode': []}
        self.decoded_strings = {'ansi': [], 'unicode': []}
        self.last_data = [0, 0]
        self.last_event = {}
        self.set_start_time()
        self.runtime = 0
        self.meta = {}
        self.runs = []

    def add_input_metadata(self, meta):
        """
        Add top level profiler fields containing metadata for the
        module that will be emulated
        """
        self.meta = meta

    def set_start_time(self):
        """
        Get the start time for a sample so we can time the execution length
        """
        self.start_time = time.time()

    def get_run_time(self):
        """
        Get the time spent emulating a specific "run"
        """
        return time.time() - self.start_time

    def stop_run_clock(self):
        """
        Stop the runtime clock to include in the report
        """
        self.runtime = self.get_run_time()

    def get_epoch_time(self):
        """
        Get the current time in epoch format
        """
        return int(time.time())

    def add_run(self, run):
        """
        Add a new run to the captured run list
        """
        self.runs.append(run)

    def handle_binary_data(self, data):
        """
        Compress and encode binary data to be included in a report
        """
        return b64encode(data).decode('utf-8')

    def log_error(self, error):
        """
        Log a top level emulator error for the emulation report
        """
        if not self.meta.get('errors'):
            self.meta['errors'] = []
        self.meta['errors'].append(error)

    def log_dropped_files(self, run, files):
        for f in files:
            data = f.get_data()
            if data is None:
                continue

            _hash = f.get_hash()
            entry = {'path': f.get_path(), 'size': len(data), 'sha256': _hash}
            run.dropped_files.append(entry)

    def log_api(self, run, pc, name, ret, argv, ctx=[]):
        """
        Log a call to an OS API. This includes arguments, return address, and return value
        """

        run.num_apis += 1

        if name not in run.unique_apis:
            run.api_hash.update(name.lower().encode('utf-8'))
            run.unique_apis.append(name)

        if not run.apis:
            run.apis = []

        pc = hex(pc)

        if ret is not None:
            ret = hex(ret)

        args = argv.copy()
        for i, arg in enumerate(args):
            if isinstance(arg, int):
                args[i] = hex(arg)

        entry = {'pc': pc, 'api_name': name, 'args': args, 'ret_val': ret}

        if entry not in run.apis[-3:]:
            run.apis.append(entry)

    def log_file_access(self, run, path, event_type, data=None,
                        handle=0, disposition=[], access=[], buffer=0,
                        size=None):
        """
        Log file access events. This will include things like handles being opened,
        data reads, and data writes.
        """
        enc = None
        if data:
            enc = self.handle_binary_data(data[:1024])

        for et in ('write', 'read'):
            if event_type == et:
                for fa in run.file_access:
                    if path == fa.get('path') and fa['event'] == et:
                        if size:
                            fa['size'] += size
                        if enc:
                            fa["data"] += enc
                        return

        event = {'event': event_type, 'path': path}
        if enc:
            event.update({'data': enc})

        if handle:
            event.update({'handle': handle})

        if size is not None:
            event.update({'size': size})

        if buffer:
            event.update({'buffer': hex(buffer)})

        if disposition:
            event.update({'open_flags': disposition})

        if access:
            event.update({'access_flags': access})

        if event not in run.file_access:
            run.file_access.append(event)

    def log_registry_access(self, run, path, event_type, value_name=None, data=None,
                            handle=0, disposition=[], access=[], buffer=0,
                            size=None):
        """
        Log registry access events. This includes values and keys being accessed and
        being read/written
        """
        enc = None
        if data:
            enc = self.handle_binary_data(data[:1024])

        event = {'event': event_type, 'path': path}
        if enc:
            event.update({'data': enc})

        if handle:
            event.update({'handle': hex(handle)})

        if value_name:
            event.update({'value_name': value_name})

        if size is not None:
            event.update({'size': size})

        if buffer:
            event.update({'buffer': hex(buffer)})

        if disposition:
            event.update({'open_flags': disposition})

        if access:
            event.update({'access_flags': access})

        if event not in run.registry_access:
            run.registry_access.append(event)

    def log_process_event(self, run, proc, event_type, kwargs):
        """
        Log events related to a process accessing another process. This includes:
        creating a child process, reading/writing to a process, or creating a thread
        within another process.
        """
        event = {}
        if event_type == PROC_CREATE:
            event.update({'event': event_type})
            event.update({'pid': proc.get_id()})
            event.update({'path': proc.get_process_path()})
            event.update({'cmdline': proc.get_command_line()})

        elif event_type == MEM_ALLOC:
            event.update({'event': event_type})
            event.update({'pid': proc.get_id()})
            event.update({'path': proc.get_process_path()})
            event.update(kwargs)

        elif event_type == MEM_PROTECT:
            event.update({'event': event_type})
            event.update({'pid': proc.get_id()})
            event.update({'path': proc.get_process_path()})
            event.update(kwargs)

        elif event_type == MEM_WRITE:
            base = kwargs['base']
            size = kwargs['size']
            data = kwargs['data']
            last_base, last_size = self.last_data
            last_evt_type = self.last_event.get('event')
            if event_type == last_evt_type and (last_base + last_size) == base:
                self.last_event['data'] += data
                self.last_event['size'] += len(data)
                self.last_data = [base, size]
                return
            event.update({'event': event_type})
            event.update({'pid': proc.get_id()})
            event.update({'path': proc.get_process_path()})
            data = kwargs['data']
            event.update({'data': data})
            event.update({'base': base})
            event.update({'size': size})
            self.last_data = [base, size]

        elif event_type == MEM_READ:
            base = kwargs['base']
            size = kwargs['size']
            data = kwargs['data']
            last_base, last_size = self.last_data
            last_evt_type = self.last_event.get('event')
            if event_type == last_evt_type and (last_base + last_size) == base:
                self.last_event['data'] += data
                self.last_event['size'] += len(data)
                self.last_data = [base, size]
                return
            event.update({'event': event_type})
            event.update({'pid': proc.get_id()})
            event.update({'path': proc.get_process_path()})
            data = kwargs['data']
            event.update({'data': data})
            event.update({'size': size})
            event.update({'base': base})
            self.last_data = [base, size]

        elif event_type == THREAD_INJECT:
            event.update({'event': event_type})
            event.update({'pid': proc.get_id()})
            event.update({'path': proc.get_process_path()})
            event.update({'start_addr': hex(kwargs['start_addr'])})
            event.update({'param': hex(kwargs['param'])})

        run.process_events.append(event)
        self.last_event = event

    def log_dns(self, run, domain, ip=''):
        """
        Log DNS name lookups for the emulation report
        """

        query = {"query": domain, "response": ip}
        if query not in run.network['dns']:
            run.network['dns'].append(query)

    def log_http(self, run, server, port, proto='http',
                 headers='', body=b'', secure=False):
        """
        Log HTTP traffic that occur during emulation
        """
        conns = run.network['traffic']

        proto = 'http'
        if secure:
            proto = 'https'

        http_conn = {'server': server, 'proto': 'tcp.%s' % (proto), 'port': port,
                     'headers': headers}
        if body:
            data = self.handle_binary_data(body[:0x3000])
            http_conn.update({'body': data})

        if http_conn not in conns:
            conns.append(http_conn)

    def log_dyn_code(self, run, tag, base, size):
        """
        Log code that is generated at runtime and then executed
        """

        if base not in run.dyn_code['base_addrs']:
            entry = {'tag': tag, 'base': hex(base), 'size': hex(size)}
            run.dyn_code['mmap'].append(entry)
            run.dyn_code['base_addrs'].add(base)

    def log_network(self, run, server, port, typ='unknown', proto='unknown', data=b'', method=''):
        """
        Log network activity for an emulation run
        """
        conns = run.network['traffic']

        conn = {'server': server, 'proto': proto, 'port': port}
        if data:
            data = self.handle_binary_data(data[:0x3000])
            conn.update({'data': data})

        if method:
            conn.update({'method': method})

        conn.update({'type': typ})

        conns.append(conn)

    def get_json_report(self):
        """
        Retrieve the execution profile for the emulator as a json string
        """
        profile = self.get_report()
        return json.dumps(profile, indent=4, sort_keys=False)

    def get_report(self):
        """
        Retrieve the execution profile for the emulator
        """
        profile = {}

        meta = self.meta
        meta.update({'report_version': __report_version__})
        meta.update({'emulation_total_runtime': round(self.runtime, 3)})
        meta.update({'timestamp': int(self.start_time)})

        # For now, we only support single file emulation
        exec_paths = []

        for r in self.runs:

            if r.ret_val is not None:
                ret = hex(r.ret_val)
            else:
                ret = None

            args = []
            for a in r.args:
                if isinstance(a, int):
                    args.append(hex(a))
                else:
                    args.append(a)

            ep = {'ep_type': r.type,
                  'start_addr': hex(r.start_addr),
                  'ep_args': args,
                  }

            if r.instr_cnt:
                ep.update({'instr_count': r.instr_cnt})

            ep.update(
                  {
                   'apihash': r.api_hash.hexdigest(),
                   'apis': r.apis,
                   'ret_val': ret,
                   'error': r.error
                  }
            )

            if r.handled_exceptions:
                ep.update({"handled_exceptions": r.handled_exceptions})

            if r.network and (r.network.get('dns', []) or
                              r.network.get('traffic', {})):
                ep.update({'network_events': r.network})

            if r.file_access:
                ep.update({'file_access': r.file_access})

            if r.registry_access:
                ep.update({'registry_access': r.registry_access})

            if r.process_events:
                for evt in r.process_events:
                    if evt.get('event') in (MEM_WRITE, MEM_READ):
                        evt['data'] = self.handle_binary_data(evt['data'][:1024])
                    if evt.get('base'):
                        evt['base'] = hex(evt['base'])
                ep.update({'process_events': r.process_events})

            if r.mem_access:
                mem_accesses = []
                for mmap, maccess in r.mem_access.items():
                    mem_accesses.append({'tag': mmap.get_tag(),
                                         'base': hex(mmap.get_base()),
                                         'reads': maccess.reads,
                                         'writes': maccess.writes,
                                         'execs': maccess.execs})

                ep.update({'mem_access': mem_accesses})

                sym_accesses = []
                for address, maccess in r.sym_access.items():
                    sym_accesses.append({'symbol': maccess.sym,
                                         'reads': maccess.reads,
                                         'writes': maccess.writes,
                                         'execs': maccess.execs})
                if sym_accesses:
                    ep.update({'sym_accesses': sym_accesses})

            if r.dyn_code:
                ep.update({'dynamic_code_segments': r.dyn_code['mmap']})

            exec_paths.append(ep)

            if r.dropped_files:
                ep.update({'dropped_files': r.dropped_files})

        if (self.strings['ansi'] or self.strings['unicode'] or
           self.decoded_strings['ansi'] or self.decoded_strings['unicode']):
           meta.update({'strings': {'static':self.strings, 'in_memory': self.decoded_strings}})  # noqa
        profile = {**profile, **meta}
        profile.update({'entry_points': exec_paths})
        return profile



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

# Unicorn specific wrappers and abstraction implemented here

import platform
import ctypes as ct
import pkg_resources

import unicorn as uc
import unicorn.unicorn
import unicorn.x86_const as u

import speakeasy.winenv.arch as arch
import speakeasy.common as common
from speakeasy.errors import EmuEngineError

_uc = unicorn.unicorn._uc
_uc.uc_hook_add = _uc.uc_hook_add
_uc.uc_hook_add.argtypes = [ct.c_void_p, ct.c_void_p, ct.c_void_p, ct.c_void_p,
                            ct.c_void_p, ct.c_uint64, ct.c_uint64]
_uc.uc_hook_add.restype = ct.c_uint32
hook_id = ct.c_void_p()

# Unicorn version we can assert on
# 1.0.2
__required_version__ = '1.0.2'


def is_platform_intel():
    mach = platform.machine()
    if mach in ('x86_64', 'i386', 'x86'):
        return True
    return False


class ToggleableHook(object):
    """
    Hook than can be toggled on/off at arbitrary times.
    """

    def __init__(self, cb):
        self.cb = cb
        self.enabled = False

    def enable(self):
        if self.enabled:
            return
        self.enabled = True

    def disable(self):
        self.enabled = False


class EmuEngine(object):
    """Wrapper class for underlying cpu emulation engines"""

    def __init__(self):
        self.name = 'unicorn'
        self.emu = None
        self.mmap = None
        self._callbacks = {}

        unicorn_ver = pkg_resources.get_distribution("unicorn").version
        assert __required_version__ == unicorn_ver, ('Requires unicorn version: %s' %
                                                     (__required_version__))

        self.regs = {
                        arch.X86_REG_EAX: u.UC_X86_REG_EAX,
                        arch.X86_REG_EBX: u.UC_X86_REG_EBX,
                        arch.X86_REG_ESP: u.UC_X86_REG_ESP,
                        arch.X86_REG_EIP: u.UC_X86_REG_EIP,
                        arch.X86_REG_EBP: u.UC_X86_REG_EBP,
                        arch.X86_REG_ECX: u.UC_X86_REG_ECX,
                        arch.X86_REG_EDX: u.UC_X86_REG_EDX,
                        arch.X86_REG_EDI: u.UC_X86_REG_EDI,
                        arch.X86_REG_ESI: u.UC_X86_REG_ESI,
                        arch.X86_REG_EFLAGS: u.UC_X86_REG_EFLAGS,
                        arch.AMD64_REG_RIP: u.UC_X86_REG_RIP,
                        arch.AMD64_REG_RAX: u.UC_X86_REG_RAX,
                        arch.AMD64_REG_RBX: u.UC_X86_REG_RBX,
                        arch.AMD64_REG_RSP: u.UC_X86_REG_RSP,
                        arch.AMD64_REG_RCX: u.UC_X86_REG_RCX,
                        arch.AMD64_REG_RDX: u.UC_X86_REG_RDX,
                        arch.AMD64_REG_RSI: u.UC_X86_REG_RSI,
                        arch.AMD64_REG_RDI: u.UC_X86_REG_RDI,
                        arch.AMD64_REG_RBP: u.UC_X86_REG_RBP,
                        arch.AMD64_REG_R8: u.UC_X86_REG_R8,
                        arch.AMD64_REG_R9: u.UC_X86_REG_R9,
                        arch.AMD64_REG_R10: u.UC_X86_REG_R10,
                        arch.AMD64_REG_R11: u.UC_X86_REG_R11,
                        arch.AMD64_REG_R12: u.UC_X86_REG_R12,
                        arch.AMD64_REG_R13: u.UC_X86_REG_R13,
                        arch.AMD64_REG_R14: u.UC_X86_REG_R14,
                        arch.AMD64_REG_R15: u.UC_X86_REG_R15,
                        arch.X86_REG_IDTR: u.UC_X86_REG_IDTR,
                        arch.X86_REG_XMM0: u.UC_X86_REG_XMM0,
                        arch.X86_REG_XMM1: u.UC_X86_REG_XMM1,
                        arch.X86_REG_XMM2: u.UC_X86_REG_XMM2,
                        arch.X86_REG_XMM3: u.UC_X86_REG_XMM3,
                        arch.X86_REG_GDTR: u.UC_X86_REG_GDTR,
                        arch.X86_REG_CS: u.UC_X86_REG_CS,
                        arch.X86_REG_ES: u.UC_X86_REG_ES,
                        arch.X86_REG_SS: u.UC_X86_REG_SS,
                        arch.X86_REG_DS: u.UC_X86_REG_DS,
                        arch.X86_REG_FS: u.UC_X86_REG_FS,
                        arch.X86_REG_GS: u.UC_X86_REG_GS,
                        arch.X86_REG_MSR: u.UC_X86_REG_MSR
        }

        self.mem_access = {
                            uc.UC_MEM_FETCH_UNMAPPED: common.INVALID_MEM_EXEC, # noqa
                            uc.UC_MEM_READ_UNMAPPED: common.INVALID_MEM_READ,
                            uc.UC_MEM_FETCH_PROT: common.INVAL_PERM_MEM_EXEC,
                            uc.UC_MEM_WRITE_PROT: common.INVAL_PERM_MEM_WRITE,
                            uc.UC_MEM_READ_PROT: common.INVAL_PERM_MEM_READ,
                            uc.UC_MEM_WRITE_UNMAPPED: common.INVALID_MEM_WRITE
        }

        self.perms = {
                    common.PERM_MEM_RWX: uc.UC_PROT_ALL,
                    common.PERM_MEM_WRITE: uc.UC_PROT_WRITE,
                    common.PERM_MEM_READ: uc.UC_PROT_READ,
                    common.PERM_MEM_RW: uc.UC_PROT_READ | uc.UC_PROT_WRITE
        }

        self.hook_types = {
                        common.HOOK_CODE: uc.UC_HOOK_CODE,
                        common.HOOK_MEM_ACCESS: uc.UC_HOOK_MEM_VALID,
                        common.HOOK_MEM_INVALID: uc.UC_HOOK_MEM_INVALID,
                        common.HOOK_MEM_PERM_EXEC: uc.UC_HOOK_MEM_FETCH_PROT,
                        common.HOOK_MEM_PERM_WRITE: uc.UC_HOOK_MEM_WRITE_PROT,
                        common.HOOK_MEM_READ: uc.UC_HOOK_MEM_READ,
                        common.HOOK_MEM_WRITE: uc.UC_HOOK_MEM_WRITE,
                        common.HOOK_INTERRUPT: uc.UC_HOOK_INTR,
                        common.HOOK_INSN: uc.UC_HOOK_INSN
        }

    def _sec_to_usec(self, sec):
        """
        Unicorn expects timeouts to be supplied in microsecond granularity
        """
        return sec * 1000000

    def init_engine(self, eng_arch, mode):
        """Initialize cpu engine"""
        if eng_arch == arch.ARCH_X86 or eng_arch == arch.ARCH_AMD64:
            _arch = uc.UC_ARCH_X86
        else:
            raise Exception('Invalid architecture')

        if mode == arch.BITS_32:
            _mode = uc.UC_MODE_32
        elif mode == arch.BITS_64:
            _mode = uc.UC_MODE_64
        else:
            raise Exception('Invalid bitness')

        self.emu = uc.Uc(_arch, _mode)

    def mem_map(self, base, size, perms=common.PERM_MEM_RWX):
        """Allocate memory in the cpu engine"""
        perm = self.perms.get(perms, uc.UC_PROT_ALL)
        return self.emu.mem_map(base, size, perm)

    def mem_unmap(self, addr, size):
        """Free memory in the cpu engine"""
        return self.emu.mem_unmap(addr, size)

    def mem_regions(self):
        """Get current memory allocations from the engine"""
        return self.emu.mem_regions()

    def mem_write(self, addr, data):
        """Write data into the address space of the engine"""
        return self.emu.mem_write(addr, data)

    def mem_read(self, addr, size):
        """Read data from the address space of the engine"""
        return self.emu.mem_read(addr, size)

    def mem_protect(self, addr, size, perms):
        """Change the memory protections for pages in the emu engine"""
        perm = self.perms.get(perms, uc.UC_PROT_ALL)
        return self.emu.mem_protect(addr, size, perm)

    def reg_write(self, reg, val):
        """Modify register values"""
        ereg = self.regs.get(reg)
        if not ereg:
            raise EmuEngineError('Unknown register: %d' % (reg))
        return self.emu.reg_write(ereg, val)

    def reg_read(self, reg):
        """Read register values"""
        ereg = self.regs.get(reg)
        if not ereg:
            raise EmuEngineError('Unknown register: %d' % (reg))
        return self.emu.reg_read(ereg)

    def stop(self):
        """Stop the emulation engine"""
        return self.emu.emu_stop()

    def start(self, addr, timeout=0, count=0):
        """Start the emulation engine"""
        if count == -1:
            count = 0

        # Unicorn expects the timeout to be in microseconds, convert it here
        timeout = self._sec_to_usec(timeout)
        return self.emu.emu_start(addr, 0xFFFFFFFF, timeout=timeout, count=count)

    def hook_add(self, addr=None, cb=None, htype=None, ctx=None, begin=1, end=0, arg1=0):
        """
        Add a callback function for a specific event type or address
        """
        hook_type = self.hook_types.get(htype)
        if not hook_type:
            raise EmuEngineError('Invalid hook type')

        handle = self.emu._uch

        # The unicorn bindings have a default python wrapper. We want to use
        # our own wrapper and don't need the extra overhead. Add callbacks directly
        # to the unicorn library here.
        if hook_type == uc.UC_HOOK_INSN:
            if arg1 == u.UC_X86_INS_IN:  # IN instruction
                cb = ct.cast(unicorn.unicorn.UC_HOOK_INSN_IN_CB(cb),
                             unicorn.unicorn.UC_HOOK_INSN_IN_CB)
            elif arg1 in (u.UC_X86_INS_SYSCALL, u.UC_X86_INS_SYSENTER):  # SYSCALL/SYSENTER
                cb = ct.cast(unicorn.unicorn.UC_HOOK_INSN_SYSCALL_CB(cb),
                             unicorn.unicorn.UC_HOOK_INSN_SYSCALL_CB)
        elif hook_type == uc.UC_HOOK_CODE:
            cb = ct.cast(unicorn.unicorn.UC_HOOK_CODE_CB(cb),
                         unicorn.unicorn.UC_HOOK_CODE_CB)
        elif hook_type in (uc.UC_HOOK_MEM_READ, uc.UC_HOOK_MEM_WRITE):
            cb = ct.cast(unicorn.unicorn.UC_HOOK_MEM_ACCESS_CB(cb),
                         unicorn.unicorn.UC_HOOK_MEM_ACCESS_CB)
        elif hook_type == uc.UC_HOOK_MEM_INVALID:
            cb = ct.cast(unicorn.unicorn.UC_HOOK_MEM_INVALID_CB(cb),
                         unicorn.unicorn.UC_HOOK_MEM_INVALID_CB)
        else:
            return self.emu.hook_add(htype=hook_type, callback=cb, user_data=ctx,
                                     begin=begin, end=end)
        ptr = ct.cast(cb, ct.c_void_p)
        # uc_hook_add requires an additional paramter for the hook type UC_HOOK_INSN
        if hook_type == uc.UC_HOOK_INSN:
            insn = ct.c_int(arg1)
            rv = _uc.uc_hook_add(handle, ct.byref(hook_id), hook_type, ptr.value,
                                 None, begin, end, insn)
        else:
            rv = _uc.uc_hook_add(handle, ct.byref(hook_id), hook_type, ptr.value,
                                 None, begin, end)
        if rv != uc.UC_ERR_OK:
            raise uc.UcError(rv)

        th = ToggleableHook(cb)
        self._callbacks.update({hook_id.value: th})

        return hook_id.value

    def hook_enable(self, hook_handle):
        """
        Enable a previously disabled hook
        """
        hook = self._callbacks.get(hook_handle)
        if hook:
            return hook.enable()

    def hook_disable(self, hook_handle):
        """
        Disable a previously enabled hook
        """
        hook = self._callbacks.get(hook_handle)
        if hook:
            return hook.disable()

    def hook_remove(self, hid):
        return self.emu.hook_del(hid)



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

from socket import inet_aton
from urllib.parse import urlparse
import speakeasy.winenv.arch as _arch

import speakeasy.windows.netman as netman
import speakeasy.winenv.defs.wininet as windefs

from .. import api


def is_ip_address(ip):
    try:
        inet_aton(ip)
        return True
    except Exception:
        return False


class WinHttp(api.ApiHandler):
    """
    Implements HTTP functions from winhttp.dll
    """

    name = 'winhttp'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):

        super(WinHttp, self).__init__(emu)

        self.funcs = {}
        self.data = {}
        self.win = None
        self.netman = netman.NetworkManager(config=emu.get_network_config())
        super(WinHttp, self).__get_hook_attrs__(self)

    @apihook('WinHttpOpen', argc=5, conv=_arch.CALL_CONV_STDCALL)
    def WinHttpOpen(self, emu, argv, ctx={}):
        """
        WINHTTPAPI HINTERNET WinHttpOpen(
          LPCWSTR pszAgentW,
          DWORD   dwAccessType,
          LPCWSTR pszProxyW,
          LPCWSTR pszProxyBypassW,
          DWORD   dwFlags
        );
        """

        ua, access, proxy, bypass, flags = argv

        if ua:
            ua = self.read_mem_string(ua, 2)
            argv[0] = ua
        if proxy:
            proxy = self.read_mem_string(proxy, 2)
            argv[2] = proxy
        if bypass:
            bypass = self.read_mem_string(bypass, 2)
            argv[3] = bypass

        conn = self.netman.new_wininet_inst(ua, access, proxy, bypass, flags)
        hnd = conn.get_handle()
        return hnd

    @apihook('WinHttpConnect', argc=4, conv=_arch.CALL_CONV_STDCALL)
    def WinHttpConnect(self, emu, argv, ctx={}):
        """
        WINHTTPAPI HINTERNET WinHttpConnect(
          IN HINTERNET     hSession,
          IN LPCWSTR       pswzServerName,
          IN INTERNET_PORT nServerPort,
          IN DWORD         dwReserved
        );
        """
        hnd, server, port, reserve = argv

        if server:
            server = self.read_mem_string(server, 2)
            argv[1] = server

        wini = self.netman.get_wininet_object(hnd)

        if not wini:
            return 0

        sess = wini.new_session(server, port, None, None,
                                0, 0, None)
        hdl = sess.get_handle()
        return hdl

    @apihook('WinHttpOpenRequest', argc=7, conv=_arch.CALL_CONV_STDCALL)
    def WinHttpOpenRequest(self, emu, argv, ctx={}):
        """
        WINHTTPAPI HINTERNET WinHttpOpenRequest(
          IN HINTERNET hConnect,
          IN LPCWSTR   pwszVerb,
          IN LPCWSTR   pwszObjectName,
          IN LPCWSTR   pwszVersion,
          IN LPCWSTR   pwszReferrer,
          IN LPCWSTR   *ppwszAcceptTypes,
          IN DWORD     dwFlags
        );
        """
        hnd, verb, objname, ver, ref, accepts, flags = argv

        if verb:
            verb = self.read_mem_string(verb, 2)
            argv[1] = verb
        if objname:
            objname = self.read_mem_string(objname, 2)
            argv[2] = objname
        if ver:
            ver = self.read_mem_string(ver, 2)
            argv[3] = ver
        if ref:
            ref = self.read_mem_string(ref, 2)
            argv[4] = ref
        if accepts:
            accepts = self.read_mem_string(accepts, 2)
            argv[5] = accepts

        defs = windefs.get_flag_defines(flags)
        argv[6] = ' | '.join(defs)

        sess = self.netman.get_wininet_object(hnd)
        req = sess.new_request(verb, objname, ver, ref, accepts, defs, None)
        hdl = req.get_handle()

        return hdl

    @apihook('WinHttpGetIEProxyConfigForCurrentUser', argc=1, conv=_arch.CALL_CONV_STDCALL)
    def WinHttpGetIEProxyConfigForCurrentUser(self, emu, argv, ctx={}):
        """
        BOOLAPI WinHttpGetIEProxyConfigForCurrentUser(
          IN OUT WINHTTP_CURRENT_USER_IE_PROXY_CONFIG *pProxyConfig
        );
        """

        proxy_config, = argv

        if proxy_config:
            self.mem_write(proxy_config, (1).to_bytes(4, 'little'))

        return True

    @apihook('WinHttpGetProxyForUrl', argc=4, conv=_arch.CALL_CONV_STDCALL)
    def WinHttpGetProxyForUrl(self, emu, argv, ctx={}):
        """
        BOOLAPI WinHttpGetProxyForUrl(
          IN HINTERNET                 hSession,
          IN LPCWSTR                   lpcwszUrl,
          IN WINHTTP_AUTOPROXY_OPTIONS *pAutoProxyOptions,
          OUT WINHTTP_PROXY_INFO       *pProxyInfo
        );
        """

        hnd, url, proxopts, proxinfo = argv

        if url:
            url = self.read_mem_string(url, 2)
            argv[1] = url

        return True

    @apihook('WinHttpSetOption', argc=4, conv=_arch.CALL_CONV_STDCALL)
    def WinHttpSetOption(self, emu, argv, ctx={}):
        """
        BOOLAPI WinHttpSendRequest(
          IN HINTERNET hRequest,
          LPCWSTR      lpszHeaders,
          IN DWORD     dwHeadersLength,
          LPVOID       lpOptional,
          IN DWORD     dwOptionalLength,
          IN DWORD     dwTotalLength,
          IN DWORD_PTR dwContext
        );
        """
        hnd, option, buff, buflen = argv

        return True

    @apihook('WinHttpSendRequest', argc=7, conv=_arch.CALL_CONV_STDCALL)
    def WinHttpSendRequest(self, emu, argv, ctx={}):
        """
        BOOLAPI WinHttpSendRequest(
          IN HINTERNET hRequest,
          LPCWSTR      lpszHeaders,
          IN DWORD     dwHeadersLength,
          LPVOID       lpOptional,
          IN DWORD     dwOptionalLength,
          IN DWORD     dwTotalLength,
          IN DWORD_PTR dwContext
        );
        """
        hnd, headers, hdrlen, lpOptional, dwOptionalLength, totlen, context = argv

        body = b''

        if headers:
            headers = self.read_mem_string(headers, 2)
            argv[1] = headers

        if lpOptional and dwOptionalLength:
            body = self.mem_read(lpOptional, dwOptionalLength)

        req = self.netman.get_wininet_object(hnd)
        srv = req.get_server()
        port = req.get_port()

        if not is_ip_address(srv):
            self.log_dns(srv)

        rv = 1
        req_str = req.format_http_request(headers=headers)

        self.log_http(srv, port, headers=req_str,
                      body=body, secure=req.is_secure())
        return rv

    @apihook('WinHttpReceiveResponse', argc=2, conv=_arch.CALL_CONV_STDCALL)
    def WinHttpReceiveResponse(self, emu, argv, ctx={}):
        """
        WINHTTPAPI BOOL WinHttpReceiveResponse(
          IN HINTERNET hRequest,
          IN LPVOID    lpReserved
        );
        """
        hnd, lpReserved = argv

        return True

    @apihook('WinHttpReadData', argc=4, conv=_arch.CALL_CONV_STDCALL)
    def WinHttpReadData(self, emu, argv, ctx={}):
        """
        BOOLAPI WinHttpReadData(
          IN HINTERNET hRequest,
          LPVOID       lpBuffer,
          IN DWORD     dwNumberOfBytesToRead,
          OUT LPDWORD  lpdwNumberOfBytesRead
        );
        """
        hnd, buf, size, bytes_read = argv

        rv = 1

        req = self.netman.get_wininet_object(hnd)
        resp = req.get_response()
        data = resp.read(size)

        if buf:
            self.mem_write(buf, data)

        if bytes_read:
            self.mem_write(bytes_read, (len(data)).to_bytes(4, 'little'))

        return rv

    @apihook('WinHttpCrackUrl', argc=4, conv=_arch.CALL_CONV_STDCALL)
    def WinHttpCrackUrl(self, emu, argv, ctx={}):
        """
        BOOLAPI WinHttpCrackUrl(
            LPCWSTR          pwszUrl,
            DWORD            dwUrlLength,
            DWORD            dwFlags,
            LPURL_COMPONENTS lpUrlComponents
        );
        """
        pwszUrl, dwUrlLength, dwFlags, lpUrlComponents = argv
        cw = 2  # Wide
        rv = False
        # TODO : implement flags
        # url = self.read_mem_string(pwszUrl, dwUrlLength)
        if pwszUrl and lpUrlComponents:
            url = self.read_mem_string(pwszUrl, cw)
            argv[0] = url
            rv = True

            uc = windefs.URL_COMPONENTS(emu.get_ptr_size())
            url_comp = self.mem_cast(uc, lpUrlComponents)

            crack = urlparse(url)
            if crack.scheme == 'https':
                url_comp.nScheme = windefs.INTERNET_SCHEME_HTTPS
            elif crack.scheme == 'http':
                url_comp.nScheme = windefs.INTERNET_SCHEME_HTTP
            if url_comp.dwHostNameLength > 0:
                if url_comp.lpszHostName:
                    host = crack.netloc + '\x00'
                    enc = self.get_encoding(cw)
                    self.mem_write(url_comp.lpszHostName, host.encode(enc))
                else:
                    offset = url.find(crack.netloc)
                    ptr = pwszUrl + (offset * cw)
                    url_comp.lpszHostName = ptr
                    url_comp.dwHostNameLength = len(crack.netloc)

            self.mem_write(lpUrlComponents, url_comp.get_bytes())

        return rv

    @apihook('WinHttpAddRequestHeaders', argc=4, conv=_arch.CALL_CONV_STDCALL)
    def WinHttpAddRequestHeaders(self, emu, argv, ctx={}):
        """
        BOOLAPI WinHttpAddRequestHeaders(
          HINTERNET hRequest,
          LPCWSTR   lpszHeaders,
          DWORD     dwHeadersLength,
          DWORD     dwModifiers
        );
        """
        hnd, headers, dwHeaderlen, dwModfier = argv

        headers = self.read_wide_string(headers, dwHeaderlen)
        argv[1] = headers
        flags = windefs.get_header_info_winhttp(dwModfier)
        argv[3] = ' | '.join(flags)
        rv = 1

        return rv

    @apihook('WinHttpQueryHeaders', argc=6, conv=_arch.CALL_CONV_STDCALL)
    def WinHttpQueryHeaders(self, emu, argv, ctx={}):
        """
       BOOLAPI WinHttpQueryHeaders(
          HINTERNET hRequest,
          DWORD     dwInfoLevel,
          LPCWSTR   pwszName,
          LPVOID    lpBuffer,
          LPDWORD   lpdwBufferLength,
          LPDWORD   lpdwIndex
        );
        """
        hnd, dwInfoLevel, name, buffer, bufferLen, index = argv

        header_query = windefs.get_header_query(dwInfoLevel)
        argv[2] = header_query

        if buffer == 0:
            emu.set_last_error(windefs.ERROR_INSUFFICIENT_BUFFER)
            return 0

        # If program checks for WINHTTP_QUERY_STATUS_CODE and the buffer is set, write '200' to buffer
        if (header_query == windefs.WINHTTP_QUERY_STATUS_CODE) and (buffer != 0):
            self.mem_write(buffer, b'\x32\x00\x30\x00\x30\x00\x00\x00')
            argv[3] = buffer
            self.mem_write(bufferLen, 8)
            argv[4] = bufferLen

        argv[5] = 0
        rv = 1

        return rv

    @apihook('WinHttpCloseHandle', argc=1, conv=_arch.CALL_CONV_STDCALL)
    def WinHttpCloseHandle(self, emu, argv, ctx={}):
        """
        BOOLAPI WinHttpCloseHandle(
          HINTERNET hInternet
        );
        """
        rv = 1

        return rv



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

from speakeasy.struct import EmuStruct, Ptr
import ctypes as ct


class NDIS_OBJECT_HEADER(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Type = ct.c_uint8
        self.Revision = ct.c_uint8
        self.Size = ct.c_uint16


class NDIS_GENERIC_OBJECT(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Header = NDIS_OBJECT_HEADER
        self.Caller = Ptr
        self.CallersCaller = Ptr
        self.DriverObject = Ptr


class NET_BUFFER_LIST_POOL_PARAMETERS(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Header = NDIS_OBJECT_HEADER
        self.ProtocolId = ct.c_uint8
        self.fAllocateNetBuffer = ct.c_uint8
        self.ContextSize = ct.c_uint16
        self.PoolTag = ct.c_uint32
        self.DataSize = ct.c_uint32


class NET_BUFFER_LIST(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Next = Ptr
        self.FirstNetBuffer = Ptr
        self.Context = Ptr
        self.ParentNetBufferList = Ptr
        self.NdisPoolHandle = Ptr
        self.NdisReserved = Ptr * 2
        self.ProtocolReserved = Ptr * 4
        self.MiniportReserved = Ptr * 2
        self.Scratch = Ptr
        self.SourceHandle = Ptr
        self.NblFlags = ct.c_uint32
        self.ChildRefCount = ct.c_uint32
        self.Flags = ct.c_uint32
        self.NetBufferListInfo = Ptr * 11


class NET_BUFFER_DATA(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Next = Ptr
        self.CurrentMdl = Ptr
        self.CurrentMdlOffset = ct.c_uint32
        self.NbDataLength = ct.c_uint32
        self.MdlChain = Ptr
        self.DataOffset = ct.c_uint32


class NET_BUFFER_HEADER(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.NetBufferData = NET_BUFFER_DATA
        self.Link = Ptr


class NET_BUFFER(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Link = Ptr
        self.NetBufferHeader = NET_BUFFER_HEADER
        self.ChecksumBias = Ptr
        self.Reserved = Ptr
        self.NdisPoolHandle = Ptr
        self.NdisReserved = Ptr * 2
        self.ProtocolReserved = Ptr * 6
        self.MiniportReserved = Ptr * 4
        self.DataPhysicalAddress = ct.c_uint64
        self.SharedMemoryInfo = Ptr



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import setuptools

with open('README.md', 'r', encoding='utf8') as fh:
    long_description = fh.read()

with open("speakeasy/version.py", encoding='utf8') as fp:
    vl = fp.readline()
    gv, ver_num = vl.split('=')
    if gv.strip() != '__version__':
        raise Exception('Invalid version file found')
    version = ver_num.strip().strip("\"\'")

setuptools.setup(
    name='speakeasy-emulator',
    author='Andrew Davis',
    description='Speakeasy malware emulation framework',
    version=version,
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=setuptools.find_packages(),
    url='https://github.com/fireeye/speakeasy',
    include_package_data=True,
    install_requires=[
        'pefile',
        'capstone',
        'lznt1',
        'unicorn==1.0.2',
        'jsonschema'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



# Copyright (C) 2021 FireEye, Inc. All Rights Reserved.

import speakeasy.winenv.defs.windows.windows as windefs
import speakeasy.winenv.defs.windows.netapi32 as netapi32defs

from .. import api


class NetApi32(api.ApiHandler):
    """
    Implements exported functions from netapi32.dll
    """
    name = 'netapi32'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):
        super(NetApi32, self).__init__(emu)
        super(NetApi32, self).__get_hook_attrs__(self)

    @apihook('NetGetJoinInformation', argc=3)
    def NetGetJoinInformation(self, emu, argv, ctx={}):
        """
        NET_API_STATUS NET_API_FUNCTION NetGetJoinInformation(
          LPCWSTR lpServer,
          LPWSTR *lpNameBuffer,
          PNETSETUP_JOIN_STATUS BufferType
        );
        """
        lpServer, lpNameBuffer, BufferType = argv

        if lpServer:
            server = self.read_wide_string(lpServer)
            argv[0] = server

        # Assumes the server being queried is the local computer
        domain = emu.get_domain()
        argv[1] = domain
        namebuf = self.mem_alloc(emu.get_ptr_size())
        self.write_wide_string(domain, namebuf)
        self.mem_write(lpNameBuffer, namebuf.to_bytes(emu.get_ptr_size(), 'little'))

        argv[2] = netapi32defs.NetSetupDomainName
        self.mem_write(BufferType, netapi32defs.NetSetupDomainName.to_bytes(4, 'little'))

        return netapi32defs.NERR_Success

    @apihook('NetWkstaGetInfo', argc=3)
    def NetWkstaGetInfo(self, emu, argv, ctx={}):
        """
        NET_API_STATUS NET_API_FUNCTION NetWkstaGetInfo(
          LMSTR  servername,
          DWORD  level,
          LPBYTE *bufptr
        );
        """
        servername, level, bufptr = argv

        if level not in [100, 101, 102]:
            return windefs.ERROR_INVALID_LEVEL

        if level == 100:
            wki = netapi32defs.WKSTA_INFO_100(emu.get_ptr_size())
        elif level == 101:
            wki = netapi32defs.WKSTA_INFO_101(emu.get_ptr_size())

            # Using empty string
            lanroot_ptr = self.mem_alloc(2)
            self.mem_write(lanroot_ptr, b'\x00\x00')
            wki.wki_lanroot = lanroot_ptr
        else:
            wki = netapi32defs.WKSTA_INFO_102(emu.get_ptr_size())

            # Using empty string
            lanroot_ptr = self.mem_alloc(2)
            self.mem_write(lanroot_ptr, b'\x00\x00')
            wki.wki_lanroot = lanroot_ptr

            wki.wki_logged_on_users = 2

        wki_addr = self.mem_alloc(wki.sizeof())
        self.mem_cast(wki, wki_addr)

        platform_id = 500  # PLATFORM_ID_NT
        wki.wki_platform_id = platform_id

        hostname = emu.get_hostname()
        computername_ptr = self.mem_alloc(2 + (len(hostname) * 2))
        # Assuming Unicode; "This string is Unicode if _WIN32_WINNT or
        # FORCE_UNICODE are defined."
        self.write_mem_string(hostname, computername_ptr, width=2)
        wki.wki_computername = computername_ptr

        domain = emu.get_domain()
        langroup_ptr = self.mem_alloc(2 + (len(domain) * 2))
        self.write_mem_string(domain, langroup_ptr, width=2)
        wki.wki_langroup = langroup_ptr

        osver = emu.get_os_version()
        wki.wki_ver_major = osver['major']
        wki.wki_ver_minor = osver['minor']

        self.mem_write(wki_addr, wki.get_bytes())
        self.mem_write(bufptr, wki_addr.to_bytes(emu.get_ptr_size(), 'little'))

        return netapi32defs.NERR_Success

    @apihook('NetApiBufferFree', argc=1)
    def NetApiBufferFree(self, emu, argv, ctx={}):
        """
        NET_API_STATUS NET_API_FUNCTION NetApiBufferFree(
          _Frees_ptr_opt_ LPVOID Buffer
        );
        """
        return netapi32defs.NERR_Success



# Copyright (C) 2021 FireEye, Inc. All Rights Reserved.

import speakeasy.winenv.defs.windows.netapi32 as netapi32defs

from .. import api


class NetUtils(api.ApiHandler):

    name = 'netutils'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):

        super(NetUtils, self).__init__(emu)
        super(NetUtils, self).__get_hook_attrs__(self)

    @apihook('NetApiBufferFree', argc=1)
    def NetApiBufferFree(self, emu, argv, ctx={}):
        """
        NET_API_STATUS NET_API_FUNCTION NetApiBufferFree(
          _Frees_ptr_opt_ LPVOID Buffer
        );
        """
        return netapi32defs.NERR_Success



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import speakeasy.winenv.defs.nt.ddk as ddk
import speakeasy.windows.kernel_mods as km


class IoManager(object):
    """
    Directs IO requests to a module handler. For example, if a user mode application
    sends an ioctl to a device this can be handled here.
    """
    def __init__(self):
        super(IoManager, self).__init__()
        self.emu_kmods = [m.DriverModule() for m in km._get_kmods()]

    def dev_ioctl(self, arch, dev, ioctl, inbuf):
        rv = ddk.STATUS_INVALID_DEVICE_REQUEST
        outbuf = b''

        # Get parent driver for you
        drv = dev.get_parent_driver()
        bn = drv.get_basename()

        # Find the emulated kernel module (if any)
        mod = [m for m in self.emu_kmods if bn == m.get_mod_name()]
        if not mod:
            rv = ddk.STATUS_INVALID_DEVICE_REQUEST
            return (rv, outbuf)
        mod = mod[0]

        return mod.ioctl(arch, ioctl, inbuf)



import os

__all__ = []

dirname = os.path.dirname(__file__)
for entry in os.listdir(dirname):
    if os.path.isfile(os.path.join(dirname, entry)):
        base, ext = os.path.splitext(entry)
        if base != '__init__' and ext == '.py':
            __all__.append(base)

del os



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import os
import shlex

import ntpath
import hashlib
import capstone as cs

import speakeasy.winenv.arch as _arch
import speakeasy.windows.common as w32common
from speakeasy.profiler import Run
from speakeasy.windows.winemu import WindowsEmulator
import speakeasy.windows.objman as objman
from speakeasy.winenv.api.winapi import WindowsApi
import speakeasy.common as common
from speakeasy.errors import Win32EmuError

from speakeasy.windows.sessman import SessionManager
from speakeasy.windows.com import COM


DLL_PROCESS_DETACH = 0
DLL_PROCESS_ATTACH = 1

MAX_EXPORTS_TO_EMULATE = 10


class Win32Emulator(WindowsEmulator):
    """
    User Mode Windows Emulator Class
    """
    def __init__(self, config, argv=[], debug=False, logger=None, exit_event=None):
        super(Win32Emulator, self).__init__(config, debug=debug, logger=logger,
                                            exit_event=exit_event)

        self.last_error = 0
        self.peb_addr = 0
        self.heap_allocs = []
        self.argv = argv
        self.sessman = SessionManager(config)
        self.com = COM(config)

    def get_argv(self):
        """
        Get command line arguments (if any) that are being passed
        to the emulated process. (e.g. main(argv))
        """
        argv0 = ''
        out = []

        if len(self.argv):
            for m in self.modules:
                pe = m[0]
                emu_path = m[2]
                if pe.is_exe():
                    argv0 = emu_path
            out = [argv0] + self.argv
        elif self.command_line:
            out = shlex.split(self.command_line, posix=False)
        return out

    def set_last_error(self, code):
        """
        Set the last error code for the current thread
        """
        if self.curr_thread:
            self.curr_thread.set_last_error(code)

    def get_last_error(self):
        """
        Get the last error code for the current thread
        """
        if self.curr_thread:
            return self.curr_thread.get_last_error()

    def get_session_manager(self):
        """
        Get the session manager for the emulator. This will manage things like desktops,
        windows, and session isolation
        """
        return self.sessman

    def add_vectored_exception_handler(self, first, handler):
        """
        Add a vectored exception handler that will be executed on an exception
        """
        self.veh_handlers.append(handler)

    def remove_vectored_exception_handler(self, handler):
        """
        Remove a vectored exception handler
        """
        self.veh_handlers.remove(handler)

    def get_processes(self):
        if len(self.processes) <= 1:
            self.init_processes(self.config_processes)
        return self.processes

    def init_processes(self, processes):
        """
        Initialize configured processes set in the emulator config
        """
        for proc in processes:
            p = objman.Process(self)
            self.add_object(p)

            p.name = proc.get('name', '')
            new_pid = proc.get('pid')
            if new_pid:
                p.pid = new_pid

            base = proc.get('base_addr')

            if isinstance(base, str):
                base = int(base, 16)
            p.base = base
            p.path = proc.get('path')
            p.session = proc.get('session', 0)
            p.image = ntpath.basename(p.path)

            self.processes.append(p)

    def load_module(self, path=None, data=None):
        """
        Load a module into the emulator space from the specified path
        """
        pe = self.load_pe(path=path, data=data, imp_id=w32common.IMPORT_HOOK_ADDR)

        if pe.arch == _arch.ARCH_X86:
            disasm_mode = cs.CS_MODE_32
        elif pe.arch == _arch.ARCH_AMD64:
            disasm_mode = cs.CS_MODE_64
        else:
            raise Win32EmuError('Unsupported architecture: %s', pe.arch)

        if not self.arch:
            self.arch = pe.arch
            self.set_ptr_size(self.arch)

        self.emu_eng.init_engine(_arch.ARCH_X86, pe.arch)

        if not self.disasm_eng:
            self.disasm_eng = cs.Cs(cs.CS_ARCH_X86, disasm_mode)

        if not data:
            file_name = os.path.basename(path) + '.exe'
            mod_name = os.path.splitext(file_name)[0]

        else:
            mod_hash = hashlib.sha256()
            mod_hash.update(data)
            mod_hash = mod_hash.hexdigest()
            mod_name = mod_hash
            file_name = '%s.exe' % (mod_name)

        self.api = WindowsApi(self)

        cd = self.get_cd()
        if not cd.endswith('\\'):
            cd += '\\'
        emu_path = cd + file_name

        if not data:
            with open(path, 'rb') as f:
                data = f.read()
        self.fileman.add_existing_file(emu_path, data)

        # Strings the initial buffer so that we can detect decoded strings later on
        if self.profiler and self.do_strings:
            self.profiler.strings['ansi'] = [a[1] for a in self.get_ansi_strings(data)]
            self.profiler.strings['unicode'] = [u[1] for u in self.get_unicode_strings(data)]

        # Set the emulated path
        emu_path = ''
        self.cd = self.get_cd()
        if self.cd:
            if not self.cd.endswith('\\'):
                self.cd += '\\'
            emu_path = self.cd + os.path.basename(file_name)

        pe.set_emu_path(emu_path)

        self.map_pe(pe, mod_name=mod_name, emu_path=emu_path)
        self.mem_write(pe.base, pe.mapped_image)

        self.setup()

        if not self.stack_base:
            self.stack_base, stack_addr = self.alloc_stack(0x12000)
        self.set_func_args(self.stack_base, self.return_hook)

        # Init imported data
        for addr, imp in pe.imports.items():
            mn, fn = imp
            mod, eh = self.api.get_data_export_handler(mn, fn)
            if eh:
                data_ptr = self.handle_import_data(mn, fn)
                sym = "%s.%s" % (mn, fn)
                self.global_data.update({addr: [sym, data_ptr]})
                self.mem_write(addr, data_ptr.to_bytes(self.get_ptr_size(),
                                                       'little'))
        return pe

    def run_module(self, module, all_entrypoints=False):
        """
        Begin emulating a previously loaded module

        Arguments:
            module: Module to emulate
        """

        if not module:
            self.stop()
            raise Win32EmuError('Module not found')

        # Check if any TLS callbacks exist, these run before the module's entry point
        tls = module.get_tls_callbacks()
        for i, cb_addr in enumerate(tls):
            base = module.get_base()
            if base < cb_addr < base + module.get_image_size():
                run = Run()
                run.start_addr = cb_addr
                run.type = 'tls_callback_%d' % (i)
                run.args = [base, DLL_PROCESS_ATTACH, 0]
                self.add_run(run)

        # Queue up the module's main entry point
        ep = module.base + module.ep

        run = Run()
        run.start_addr = ep

        main_exe = None
        if not module.is_exe():
            run.args = [module.base, DLL_PROCESS_ATTACH, 0]
            run.type = 'dll_entry.DLL_PROCESS_ATTACH'
            container = self.init_container_process()
            if container:
                self.processes.append(container)
                self.curr_process = container
        else:
            run.type = 'module_entry'
            main_exe = module
            run.args = [self.mem_map(8, tag='emu.module_arg_%d' % (i)) for i in range(4)]

        if main_exe:
            self.user_modules = [main_exe] + self.user_modules

        self.add_run(run)

        if all_entrypoints:
            # Only emulate a subset of all the exported functions
            # There are some modules (such as the windows kernel) with
            # thousands of exports
            exports = [k for k in module.get_exports()[: MAX_EXPORTS_TO_EMULATE]]

            if exports:
                args = [self.mem_map(8, tag='emu.export_arg_%d' % (i), base=0x41420000) for i in range(4)] # noqa
                for exp in exports:
                    if exp.name in ('DllMain', 'ServiceMain'):
                        continue
                    run = Run()
                    if exp.name:
                        fn = exp.name
                    else:
                        fn = 'no_name'

                    run.type = 'export.%s' % (fn)
                    run.start_addr = exp.address
                    # Here we set dummy args to pass into the export function
                    run.args = args
                    # Store these runs and only queue them before the unload
                    # routine this is because some exports may not be ready to
                    # be called yet
                    self.add_run(run)

        # Create an empty process object for the module if none is
        # supplied
        if len(self.processes) == 0:
            p = objman.Process(self, path=module.get_emu_path(), base=module.base,
                               pe=module, cmdline=self.command_line)
            self.curr_process = p
            self.om.objects.update({p.address: p})
            mm = self.get_address_map(module.base)
            if mm:
                mm.process = self.curr_process

        t = objman.Thread(self,
                          stack_base=self.stack_base,
                          stack_commit=module.stack_commit)

        self.om.objects.update({t.address: t})
        self.curr_process.threads.append(t)
        self.curr_thread = t

        peb = self.alloc_peb(self.curr_process)

        # Set the TEB
        self.init_teb(t, peb)

        # Begin emulation
        self.start()

    def emulate_module(self, path):
        """
        Load and emulate binary from the given path
        """
        mod = self.load_module(path)
        self.run_module(mod)

    def load_shellcode(self, path, arch, data=None):
        """
        Load position independent code (i.e. shellcode) to prepare for emulation
        """
        sc_hash = None

        if arch == 'x86':
            arch = _arch.ARCH_X86
        elif arch in ('x64', 'amd64'):
            arch = _arch.ARCH_AMD64

        self.arch = arch

        if data:
            sc_hash = hashlib.sha256()
            sc_hash.update(data)
            sc_hash = sc_hash.hexdigest()
            sc = data
        else:
            with open(path, 'rb') as scpath:
                sc = scpath.read()

            sc_hash = hashlib.sha256()
            sc_hash.update(sc)
            sc_hash = sc_hash.hexdigest()

        if self.arch == _arch.ARCH_X86:
            disasm_mode = cs.CS_MODE_32
        elif self.arch == _arch.ARCH_AMD64:
            disasm_mode = cs.CS_MODE_64
        else:
            raise Win32EmuError('Unsupported architecture: %s' % self.arch)

        self.emu_eng.init_engine(_arch.ARCH_X86, self.arch)

        if not self.disasm_eng:
            self.disasm_eng = cs.Cs(cs.CS_ARCH_X86, disasm_mode)

        sc_tag = 'emu.shellcode.%s' % (sc_hash)

        # Map the shellcode into memory
        sc_addr = self.mem_map(len(sc), tag=sc_tag)
        self.mem_write(sc_addr, sc)

        self.pic_buffers.append((path, sc_addr, len(sc)))

        sc_arch = 'unknown'
        if arch == _arch.ARCH_AMD64:
            sc_arch = 'x64'
        elif arch == _arch.ARCH_X86:
            sc_arch = 'x86'

        if self.profiler:
            self.input = {'path': path, 'sha256': sc_hash, 'size': len(sc),
                          'arch': sc_arch, 'mem_tag': sc_tag,
                          'emu_version': self.get_emu_version(),
                          'os_run': self.get_osver_string()}
            self.profiler.add_input_metadata(self.input)
            # Strings the initial buffer so that we can detect decoded strings later on
            if self.do_strings:
                self.profiler.strings['ansi'] = [a[1] for a in self.get_ansi_strings(sc)]
                self.profiler.strings['unicode'] = [u[1] for u in self.get_unicode_strings(sc)]
        self.setup()

        return sc_addr

    def run_shellcode(self, sc_addr, offset=0):
        """
        Begin emulating position independent code (i.e. shellcode) to prepare for emulation
        """

        target = None
        for sc_path, _sc_addr, size in self.pic_buffers:
            if _sc_addr == sc_addr:
                target = _sc_addr
                break

        if not target:
            raise Win32EmuError('Invalid shellcode address')

        stack_commit = 0x4000

        self.stack_base, stack_addr = self.alloc_stack(stack_commit)
        self.set_func_args(self.stack_base, self.return_hook, 0x7000)

        run = Run()
        run.type = 'shellcode'
        run.start_addr = sc_addr + offset
        run.instr_cnt = 0
        args = [self.mem_map(1024, tag='emu.shellcode_arg_%d' % (i), base=0x41420000 + i)
                for i in range(4)]
        run.args = (args)

        self.reg_write(_arch.X86_REG_ECX, 1024)

        self.add_run(run)

        # Create an empty process object for the shellcode if none is
        # supplied
        container = self.init_container_process()
        if container:
            self.processes.append(container)
            self.curr_process = container
        else:
            p = objman.Process(self)
            self.processes.append(p)
            self.curr_process = p

        mm = self.get_address_map(sc_addr)
        if mm:
            mm.set_process(self.curr_process)

        t = objman.Thread(self,
                          stack_base=self.stack_base, stack_commit=stack_commit)
        self.om.objects.update({t.address: t})
        self.curr_process.threads.append(t)

        self.curr_thread = t

        peb = self.alloc_peb(self.curr_process)

        # Set the TEB
        self.init_teb(t, peb)

        self.start()

    def alloc_peb(self, proc):
        """
        Allocate memory for the Process Environment Block (PEB)
        """

        if proc.is_peb_active:
            return
        size = proc.get_peb_ldr().sizeof()
        res, size = self.get_valid_ranges(size)
        self.mem_reserve(size, base=res, tag='emu.struct.PEB_LDR_DATA')
        proc.set_peb_ldr_address(res)

        peb = proc.get_peb()
        proc.is_peb_active = True
        peb.object.ImageBaseAddress = proc.base
        peb.object.OSMajorVersion = self.osversion['major']
        peb.object.OSMinorVersion = self.osversion['minor']
        peb.object.OSBuildNumber = self.osversion['build']
        peb.write_back()
        return peb

    def set_unhandled_exception_handler(self, handler_addr):
        """
        Establish a handler for unhandled exceptions that occur during emulation
        """
        self.unhandled_exception_filter = handler_addr

    def setup(self, stack_commit=0):

        # Set the emulator to run in protected mode
        self.om = objman.ObjectManager(emu=self)

        arch = self.get_arch()
        self._setup_gdt(arch)
        self.setup_user_shared_data()
        self.set_ptr_size(self.arch)

        if arch == _arch.ARCH_X86:
            self.peb_addr = self.fs_addr + 0x30
        elif arch == _arch.ARCH_AMD64:
            self.peb_addr = self.gs_addr + 0x60

        self.api = WindowsApi(self)

        # Init symlinks
        for sl in self.symlinks:
            self.om.add_symlink(sl['name'], sl['target'])

        self.init_sys_modules(self.config_system_modules)

    def init_sys_modules(self, modules_config):
        """
        Get the system modules (e.g. drivers) that are loaded in the emulator
        """
        sys_mods = []

        for modconf in modules_config:

            mod = w32common.DecoyModule()
            mod.name = modconf['name']
            base = modconf.get('base_addr')
            if isinstance(base, str):
                base = int(base, 16)

            mod.decoy_base = base
            mod.decoy_path = modconf['path']

            drv = modconf.get('driver')
            if drv:
                devs = drv.get('devices')
                for dev in devs:
                    name = dev.get('name', '')
                    do = self.new_object(objman.Device)
                    do.name = name

            sys_mods.append(mod)
        return sys_mods

    def init_container_process(self):
        """
        Create a process to be used to host shellcode or DLLs
        """
        for p in self.config_processes:
            if p.get('is_main_exe'):
                name = p.get('name', '')
                emu_path = p.get('path', '')
                base = p.get('base_addr', 0)
                if isinstance(base, str):
                    base = int(base, 0)
                cmd_line = p.get('command_line', '')

                proc = objman.Process(self, name=name,
                                      path=emu_path, base=base, cmdline=cmd_line)
                return proc
        return None

    def get_user_modules(self):
        """
        Get the user modules (e.g. dlls) that are loaded in the emulator
        """
        # Generate the decoy user module list
        if len(self.user_modules) < 2:
            # Check if we have a host process configured
            proc_mod = None
            for p in self.config_processes:

                if not self.user_modules and p.get('is_main_exe'):
                    proc_mod = p
                    break

            if proc_mod:
                all_user_mods = [proc_mod] + self.config_user_modules
                user_modules = self.init_user_modules(all_user_mods)
            else:
                user_modules = self.init_user_modules(self.config_user_modules)

            self.user_modules += user_modules
            # add sample to user modules list if it is a dll
            if self.modules and not self.modules[0][0].is_exe():
                self.user_modules.append(self.modules[0][0])

        return self.user_modules

    def exit_process(self):
        """
        An emulated binary is attempted to terminate its current process.
        Signal that the run has finished.
        """
        self.enable_code_hook()
        self.run_complete = True

    def _hook_mem_unmapped(self, emu, access, address, size, value, user_data):

        _access = self.emu_eng.mem_access.get(access)

        if _access == common.INVALID_MEM_READ:
            p = self.get_current_process()
            pld = p.get_peb_ldr()
            if address > pld.address and address < (pld.address + pld.sizeof()):
                self.mem_map_reserve(pld.address)
                user_mods = self.get_user_modules()
                self.init_peb(user_mods)
                return True
        return super(Win32Emulator, self)._hook_mem_unmapped(emu, access, address, size,
                                                             value, user_data)

    def set_hooks(self):
        """Set the emulator callbacks"""

        super(Win32Emulator, self).set_hooks()

        if not self.builtin_hooks_set:
            self.add_mem_invalid_hook(cb=self._hook_mem_unmapped)
            self.add_interrupt_hook(cb=self._hook_interrupt)
            self.builtin_hooks_set = True

        self.set_mem_tracing_hooks()

    def stop(self):
        self.run_complete = True
        super(Win32Emulator, self).stop()

    def on_emu_complete(self):
        """
        Called when all runs have completed emulation
        """
        if not self.emu_complete:
            self.emu_complete = True
            if self.do_strings and self.profiler:
                dec_ansi, dec_unicode = self.get_mem_strings()
                dec_ansi = [a[1] for a in dec_ansi if a not in self.profiler.strings['ansi']]
                dec_unicode = [u[1] for u in dec_unicode
                               if u not in self.profiler.strings['unicode']]
                self.profiler.decoded_strings['ansi'] = dec_ansi
                self.profiler.decoded_strings['unicode'] = dec_unicode
        self.stop()

    def on_run_complete(self):
        """
        Clean up after a run completes. This function will pop the
        next run from the run queue and emulate it.
        """
        self.run_complete = True
        self.curr_run.ret_val = self.get_return_val()
        if self.profiler:
            self.profiler.log_dropped_files(self.curr_run, self.get_dropped_files())

        return self._exec_next_run()

    def heap_alloc(self, size, heap='None'):
        """
        Allocate a memory chunk and add it to the "heap"
        """
        addr = self.mem_map(size, base=None, tag='api.heap.%s' % (heap))
        self.heap_allocs.append((addr, size, heap))
        return addr
